"""Module with high-level operations to segment images."""

import logging
import math
import os
from pathlib import Path
from typing import Optional

import numpy as np

# import keras as kr
import pandas as pd
import tensorflow as tf
from PIL import Image
from tensorflow import keras as kr

import orthoseg.model.model_factory as mf
import orthoseg.model.model_helper as mh

# Get a logger...
logger = logging.getLogger(__name__)


def train(
    traindata_dir: Path,
    validationdata_dir: Path,
    model_save_dir: Path,
    segment_subject: str,
    traindata_id: int,
    hyperparams: mh.HyperParams,
    model_preload_filepath: Optional[Path] = None,
    image_width: int = 512,
    image_height: int = 512,
    image_subdir: str = "image",
    mask_subdir: str = "mask",
    save_augmented_subdir: Optional[str] = None,
):
    """Create/load a neural network and train it.

    Data from the train and validation directories specified for the training.

    The best models will be saved to model_save_dir. The filenames of the
    models will be constructed like this:
    {model_save_base_filename}_{combined_acc}_{train_acc}_{validation_acc}_{epoch}
        * combined_acc: average of train_acc and validation_acc
        * train_acc: the jaccard coëficient of train dataset for the model
        * validation_acc: the jaccard coëficient of validation dataset
    In the scripts, if the "best model" is mentioned, this is the one with the
    highest "combined_acc".

    Args:
        traindata_dir: dir where the train data is located
        validationdata_dir: dir where the validation data is located
        model_save_dir: dir where (intermediate) best models will be saved
        segment_subject (str): segment subject
        traindata_id (int): train data version
        hyperparams (mh.HyperParams): the hyper parameters to use for the model
        model_preload_filepath: filepath to the model to continue training on,
            or None if you want to start from scratch
        image_width: width the input images will be rescaled to for training
        image_height: height the input images will be rescaled to for training
        image_subdir: subdir where the images can be found in traindata_dir and
            validationdata_dir
        mask_subdir: subdir where the corresponding masks can be found in traindata_dir
            and validationdata_dir
        save_augmented_subdir: (str, optional): the subdirectory to save the augmented
            images to. If None, the images aren't saved. Defaults to None.
    """
    # These are the augmentations that will be applied to the input training
    # images/masks.
    # Remark: fill_mode + cval are defined as they are so missing pixels after
    #    eg. rotation are filled with 0, and so the mask will take care that they are
    #    +- ignored.

    # Create the train generator
    train_gen = create_train_generator(
        input_data_dir=traindata_dir,
        image_subdir=image_subdir,
        mask_subdir=mask_subdir,
        image_augment_dict=hyperparams.train.image_augmentations,
        mask_augment_dict=hyperparams.train.mask_augmentations,
        batch_size=hyperparams.train.batch_size,
        target_size=(image_width, image_height),
        nb_classes=len(hyperparams.architecture.classes),
        save_to_subdir=save_augmented_subdir,
        seed=2,
    )

    # Create validation generator
    validation_augmentations = {
        "rescale": hyperparams.train.image_augmentations["rescale"]
    }
    validation_mask_augmentations = {
        "rescale": hyperparams.train.mask_augmentations["rescale"]
    }
    validation_gen = create_train_generator(
        input_data_dir=validationdata_dir,
        image_subdir=image_subdir,
        mask_subdir=mask_subdir,
        image_augment_dict=validation_augmentations,
        mask_augment_dict=validation_mask_augmentations,
        batch_size=hyperparams.train.batch_size,
        target_size=(image_width, image_height),
        nb_classes=len(hyperparams.architecture.classes),
        save_to_subdir=save_augmented_subdir,
        shuffle=False,
        seed=3,
    )

    # Get the max epoch number from the log file if it exists...
    start_epoch = 0
    model_save_base_filename = mh.format_model_basefilename(
        segment_subject=segment_subject,
        traindata_id=traindata_id,
        architecture_id=hyperparams.architecture.architecture_id,
        trainparams_id=hyperparams.train.trainparams_id,
    )
    csv_log_filepath = model_save_dir / (model_save_base_filename + "_log.csv")
    if csv_log_filepath.exists() and os.path.getsize(csv_log_filepath) > 0:
        if not model_preload_filepath:
            message = (
                "log file exists but preload model file not specified: "
                f"{csv_log_filepath}"
            )
            logger.critical(message)
            raise Exception(message)

        train_log_df = pd.read_csv(csv_log_filepath, sep=";", usecols=["epoch", "lr"])
        assert isinstance(train_log_df, pd.DataFrame)
        logger.debug(f"train_log csv contents:\n{train_log_df}")
        start_epoch = train_log_df["epoch"].max()
        hyperparams.train.optimizer_params["learning_rate"] = float(
            pd.to_numeric(train_log_df["lr"], downcast="float").min()
        )
    logger.info(
        f"start_epoch: {start_epoch}, learning_rate: "
        f"{hyperparams.train.optimizer_params['learning_rate']}"
    )

    # If no existing model provided, create it from scratch
    if not model_preload_filepath:
        # If the encode should be frozen for some epochs...
        if hyperparams.train.nb_epoch_with_freeze > 0:
            freeze = True
        else:
            freeze = False

        # Get the model we want to use
        model = mf.get_model(
            architecture=hyperparams.architecture.architecture,
            nb_channels=hyperparams.architecture.nb_channels,
            nb_classes=len(hyperparams.architecture.classes),
            activation=hyperparams.architecture.activation_function,
            freeze=freeze,
        )

        # Save the model architecture to json
        model_json_filepath = model_save_dir / f"{model_save_base_filename}_model.json"
        if not model_save_dir.exists():
            model_save_dir.mkdir(parents=True)
        if not model_json_filepath.exists():
            with model_json_filepath.open("w") as dst:
                dst.write(str(model.to_json()))
    else:
        # If a preload model is provided, load that if it exists...
        if not model_preload_filepath.exists():
            message = (
                f"Error: preload model file doesn't exist: {model_preload_filepath}"
            )
            logger.critical(message)
            raise Exception(message)

        # Load the existing model
        # Remark: compiling during load crashes, so compile 'manually'
        logger.info(f"Load model from {model_preload_filepath}")
        model = mf.load_model(model_preload_filepath, compile=False)

    # Now prepare the model for training
    nb_gpu = len(tf.config.experimental.list_physical_devices("GPU"))

    # TODO: because of bug in tensorflow 1.14, multi GPU doesn't work (this way),
    # so always use standard model
    if nb_gpu <= 1:
        model_for_train = mf.compile_model(
            model=model,
            optimizer=hyperparams.train.optimizer,
            optimizer_params=hyperparams.train.optimizer_params,
            loss=hyperparams.train.loss_function,
            class_weights=hyperparams.train.class_weights,
        )
        logger.info(f"Train using single GPU or CPU, with nb_gpu: {nb_gpu}")
    else:
        # If multiple GPU's available, should create a multi-GPU model
        model_for_train = model
        logger.info(f"Train using all GPU's, with nb_gpu: {nb_gpu}")
        logger.warning("MULTI GPU TRAINING NOT TESTED BUT WILL BE TRIED ANYWAY")
        strategy = tf.distribute.MirroredStrategy()
        with strategy.scope():
            model_for_train = mf.compile_model(
                model=model,
                optimizer=hyperparams.train.optimizer,
                optimizer_params=hyperparams.train.optimizer_params,
                loss=hyperparams.train.loss_function,
                class_weights=hyperparams.train.class_weights,
            )

    # Define some callbacks for the training
    train_callbacks = []
    # Reduce the learning rate if the loss doesn't improve anymore
    reduce_lr = kr.callbacks.ReduceLROnPlateau(
        monitor="loss",
        factor=0.2,
        patience=20,
        min_lr=1e-20,
        verbose=True,
    )
    train_callbacks.append(reduce_lr)

    # Custom callback that saves the best models using both train and
    # validation metric
    # Remark: the save of the model should be done on the standard model, not
    #         on the parallel_model, otherwise issues to use it afterwards
    if nb_gpu > 1:
        model_template_for_save = model
    else:
        model_template_for_save = None
    model_checkpoint_saver = mh.ModelCheckpointExt(
        model_save_dir=model_save_dir,
        segment_subject=segment_subject,
        traindata_id=traindata_id,
        architecture_id=hyperparams.architecture.architecture_id,
        trainparams_id=hyperparams.train.trainparams_id,
        monitor_metric=hyperparams.train.monitor_metric,
        monitor_metric_mode=hyperparams.train.monitor_metric_mode,
        save_format=hyperparams.train.save_format,
        save_best_only=hyperparams.train.save_best_only,
        save_min_accuracy=hyperparams.train.save_min_accuracy,
        model_template_for_save=model_template_for_save,
    )
    train_callbacks.append(model_checkpoint_saver)

    # Callbacks for logging
    if hyperparams.train.log_tensorboard is True:
        tensorboard_log_dir = model_save_dir / (
            model_save_base_filename + "_tensorboard_log"
        )
        tensorboard_logger = kr.callbacks.TensorBoard(log_dir=str(tensorboard_log_dir))
        train_callbacks.append(tensorboard_logger)
    if hyperparams.train.log_csv is True:
        csv_logger = kr.callbacks.CSVLogger(
            str(csv_log_filepath), append=True, separator=";"
        )
        train_callbacks.append(csv_logger)

    # Stop if no more improvement
    early_stopping = kr.callbacks.EarlyStopping(
        monitor=hyperparams.train.earlystop_monitor_metric,
        patience=hyperparams.train.earlystop_patience,
        mode=hyperparams.train.earlystop_monitor_metric_mode,
        restore_best_weights=False,
    )
    train_callbacks.append(early_stopping)

    # Prepare the parameters to pass to fit...
    # Supported filetypes to train/validate on
    input_ext = [".tif", ".jpg", ".png"]

    # Calculate the size of the input datasets
    # train_dataset_size = len(glob.glob(f"{traindata_dir}/{image_subdir}/*.*"))
    train_dataset_size = 0
    for input_ext_cur in input_ext:
        traindata_image_dir = traindata_dir / image_subdir
        train_dataset_size += len(list(traindata_image_dir.rglob("*" + input_ext_cur)))
    validation_dataset_size = 0
    for input_ext_cur in input_ext:
        validationdata_image_dir = validationdata_dir / image_subdir
        validation_dataset_size += len(
            list(validationdata_image_dir.rglob("*" + input_ext_cur))
        )

    # Calculate the number of steps within an epoch
    # Remark: number of steps per epoch should be at least 1, even if nb samples
    # < batch size...
    train_steps_per_epoch = math.ceil(train_dataset_size / hyperparams.train.batch_size)
    validation_steps_per_epoch = math.ceil(
        validation_dataset_size / hyperparams.train.batch_size
    )

    # Start training
    logger.info(
        f"Start training with batch_size: {hyperparams.train.batch_size}, "
        f"train_dataset_size: {train_dataset_size}, train_steps_per_epoch: "
        f"{train_steps_per_epoch}, validation_dataset_size: {validation_dataset_size}, "
        f"validation_steps_per_epoch: {validation_steps_per_epoch}"
    )

    logger.info(f"{hyperparams.toJSON()}")
    hyperparams_filepath = (
        model_save_dir / f"{model_save_base_filename}_hyperparams.json"
    )
    hyperparams_filepath.write_text(hyperparams.toJSON())

    try:
        # If the encoder should be frozen for the first epochs, do so
        if hyperparams.train.nb_epoch_with_freeze > 0:
            logger.info(
                f"First train for {hyperparams.train.nb_epoch_with_freeze} epochs with "
                "frozen layers"
            )
            model_for_train.fit(
                train_gen,
                steps_per_epoch=train_steps_per_epoch,
                epochs=hyperparams.train.nb_epoch_with_freeze,
                validation_data=validation_gen,
                # Number of items in validation/batch_size
                validation_steps=validation_steps_per_epoch,
                callbacks=train_callbacks,
                initial_epoch=start_epoch,
                verbose=2,
            )
            mf.set_trainable(model=model_for_train, recompile=False)
            model_for_train = mf.compile_model(
                model=model_for_train,
                optimizer=hyperparams.train.optimizer,
                optimizer_params=hyperparams.train.optimizer_params,
                loss=hyperparams.train.loss_function,
                class_weights=hyperparams.train.class_weights,
            )

        # Train!
        model_for_train.fit(
            train_gen,
            steps_per_epoch=train_steps_per_epoch,
            epochs=hyperparams.train.nb_epoch,
            validation_data=validation_gen,
            # Number of items in validation/batch_size
            validation_steps=validation_steps_per_epoch,
            callbacks=train_callbacks,
            initial_epoch=start_epoch,
            verbose=2,
        )

        # Write some reporting
        train_report_path = model_save_dir / (model_save_base_filename + "_report.pdf")
        train_log_df = pd.read_csv(csv_log_filepath, sep=";")
        assert isinstance(train_log_df, pd.DataFrame)
        columns_to_keep = []
        for column in train_log_df.columns:
            if column.endswith(("accuracy", "f1-score")):
                columns_to_keep.append(column)

        train_log_vis_df = train_log_df[columns_to_keep]
        fig = train_log_vis_df.plot().get_figure()
        if fig is not None:
            fig.savefig(str(train_report_path))

    finally:
        # Release the memory from the GPU...
        # from keras import backend as K
        # K.clear_session()
        kr.backend.clear_session()


def create_train_generator(
    input_data_dir: Path,
    image_subdir: str,
    mask_subdir: str,
    image_augment_dict: dict,
    mask_augment_dict: dict,
    batch_size: int = 32,
    image_color_mode: str = "rgb",
    mask_color_mode: str = "grayscale",
    save_to_subdir: Optional[str] = None,
    image_save_prefix: str = "image",
    mask_save_prefix: str = "mask",
    nb_classes: int = 1,
    target_size: tuple[int, int] = (256, 256),
    shuffle: bool = True,
    seed: int = 1,
):
    """Creates a generator to generate and augment train images.

    The augmentations
    specified in aug_dict will be applied. For the augmentations that can be
    specified in aug_dict look at the documentation of
    keras.preprocessing.image.ImageDataGenerator

    For more info about the other parameters, check keras flow_from_directory.

    Remarks: * use the same seed for image_datagen and mask_datagen to ensure
               the transformation for image and mask is the same
             * set save_to_dir = "your path" to check results of the generator
    """
    # Init
    # Do some checks on the augmentations specified, as it is easy to
    # introduce illogical values
    if (
        image_augment_dict is not None
        and mask_augment_dict is None
        or image_augment_dict is not None
        and mask_augment_dict is None
    ):
        logger.warning(
            "Only augmentations specified for either image or mask: "
            f"image_augment_dict: {image_augment_dict}, mask_augment_dict: "
            f"{mask_augment_dict}"
        )

    # Checks that involve comparing augmentations between the image and the mask
    if image_augment_dict is not None and mask_augment_dict is not None:
        # If an augmentation is specified for image, it should be specified
        # for the mask as well and the other way around to evade issues
        for augmentation in image_augment_dict:
            if augmentation not in mask_augment_dict:
                raise Exception(
                    f"{augmentation} in image_augment_dict but not in mask_augment_dict"
                )
        for augmentation in mask_augment_dict:
            if augmentation not in image_augment_dict:
                raise Exception(
                    f"{augmentation} in mask_augment_dict but not in image_augment_dict"
                )

        # Check if the brightness range has valid values
        image_brightness_range = image_augment_dict.get("brightness_range")
        mask_brightness_range = mask_augment_dict.get("brightness_range")
        if image_brightness_range is None and mask_brightness_range is not None:
            # If brightness_range None for image, it should be None for the mask as well
            raise Exception(
                "augmentation brightness_range is None (null) in image_augment_dict "
                f"but isn't in mask_augment_dict: {mask_brightness_range}"
            )
        elif image_brightness_range is not None and mask_brightness_range is None:
            # If brightness_range None for mask, it should be None for the image as well
            raise Exception(
                "augmentation brightness_range is None (null) in mask_augment_dict but "
                f"isn't in image_augment_dict: {image_brightness_range}"
            )
        elif image_brightness_range is not None and mask_brightness_range is not None:
            # The brightness_range values should be >= 0
            if image_brightness_range[0] < 0 or image_brightness_range[1] < 0:
                raise Exception(
                    "augmentation brightness_range values should be > 0: value 1.0 = "
                    "don't do anything, 0 = black"
                )
            # brightness_range is applied to the image, it should be [1,1] for the mask
            if mask_brightness_range[0] != 1.0 or mask_brightness_range[1] != 1.0:
                raise ValueError(
                    "augmentation brightness_range is specified on the image: then the "
                    f"mask should get range [1, 1], not {mask_brightness_range}"
                )

    # Some checks specific for the mask
    if mask_augment_dict is not None:
        # Check cval value
        if "cval" in mask_augment_dict and mask_augment_dict["cval"] != 0:
            logger.warning(
                "cval typically should be 0 for the mask, even if it is different for "
                "the image, as the cval of the mask refers to these locations being of "
                f"class 'background'. It is: {mask_augment_dict['cval']}"
            )

        # If there are more than two classes, the mask will have integers as values
        # to code the different masks in, and one hot-encoding will be applied to
        # it, so it should not be rescaled!!!
        if (
            nb_classes > 2
            and "rescale" in mask_augment_dict
            and mask_augment_dict["rescale"] != 1
        ):
            raise Exception(
                f"With nb_classes > 2 ({nb_classes}), the mask should have a rescale "
                f"value of 1, not {mask_augment_dict['rescale']}"
            )

    # Now create the generators
    # Create the image generators with the augment info

    # TODO: brightness_range is buggy on 2021-08-11, with keras preprocessing
    # version 1.1, so is implemented here in a hacky way!
    image_augment_dict_temp = {
        key: value
        for (key, value) in image_augment_dict.items()
        if key != "brightness_range"
    }
    mask_augment_dict_temp = {
        key: value
        for (key, value) in mask_augment_dict.items()
        if key != "brightness_range"
    }
    image_datagen = kr.preprocessing.image.ImageDataGenerator(**image_augment_dict_temp)
    mask_datagen = kr.preprocessing.image.ImageDataGenerator(**mask_augment_dict_temp)

    # Format save_to_dir
    # Remark: flow_from_directory doesn't support Path, so supply str immediately as
    # well, otherwise, if str(Path) is used later on, it becomes 'None' instead of None!
    save_to_dir = None
    if save_to_subdir is not None:
        save_to_dir = input_data_dir / save_to_subdir
        if not save_to_dir.exists():
            save_to_dir.mkdir(parents=True)

    image_generator = image_datagen.flow_from_directory(
        directory=str(input_data_dir),
        classes=[image_subdir],
        class_mode=None,
        color_mode=image_color_mode,
        target_size=target_size,
        batch_size=batch_size,
        save_to_dir=None,
        save_prefix=image_save_prefix,
        shuffle=shuffle,
        seed=seed,
    )

    mask_generator = mask_datagen.flow_from_directory(
        directory=str(input_data_dir),
        classes=[mask_subdir],
        class_mode=None,
        color_mode=mask_color_mode,
        target_size=target_size,
        batch_size=batch_size,
        save_to_dir=None,
        save_prefix=mask_save_prefix,
        shuffle=shuffle,
        seed=seed,
    )

    train_generator = zip(image_generator, mask_generator)

    for batch_id, (image, mask) in enumerate(train_generator):
        # Cast to arrays to evade type errors
        image = np.array(image)
        mask = np.array(mask)

        # TODO: brightness_range is buggy on 2021-08-11, with keras preprocessing
        # version 1.1, so is implemented here in a hacky way!
        if (
            "brightness_range" in image_augment_dict
            and image_augment_dict["brightness_range"] is not None
            and (
                image_augment_dict["brightness_range"][0] != 1
                or image_augment_dict["brightness_range"][1] != 1
            )
        ):
            # Random brightness shift to apply to all images in batch
            brightness_shift = np.random.uniform(
                image_augment_dict["brightness_range"][0],
                image_augment_dict["brightness_range"][1],
            )
            image = image * brightness_shift

        # One-hot encode mask if multiple classes
        if nb_classes > 1:
            mask = kr.utils.to_categorical(mask, nb_classes)

        # Because the default save_to_dir option doesn't support saving the
        # augmented masks in seperate files per class, implement this here.
        if save_to_dir is not None:
            # Save mask for every class seperately
            # Get the number of images in this batch + the nb classes
            mask_shape = mask.shape
            nb_images = mask_shape[0]
            nb_classes = mask_shape[3]

            # Loop through images in this batch
            for image_id in range(nb_images):
                # Slice the next image from the array
                image_to_save = image[image_id, :, :, :]

                # Reverse the rescale if there is one
                if (
                    image_augment_dict is not None
                    and "rescale" in image_augment_dict
                    and image_augment_dict["rescale"] != 1
                ):
                    image_to_save = image_to_save / image_augment_dict["rescale"]

                # Now convert to uint8 image and save!
                # , image_color_mode
                colormode = None
                if image_color_mode.upper() == "RGB":
                    colormode = "RGB"
                im = Image.fromarray(image_to_save.astype(np.uint8), mode=colormode)
                image_dir = save_to_dir / f"{batch_id:0>4}"
                image_dir.mkdir(parents=True, exist_ok=True)
                image_path = image_dir / f"{image_id}_image.jpg"
                im.save(image_path)

                # Loop through the masks for each class
                for channel_id in range(nb_classes):
                    mask_to_save = mask[image_id, :, :, channel_id]
                    mask_dir = save_to_dir / f"{batch_id:0>4}"
                    mask_path = mask_dir / f"{image_id}_mask_{channel_id}.png"
                    im = Image.fromarray((mask_to_save * 255).astype(np.uint8))
                    im.save(mask_path)

        yield (image, mask)


# If the script is ran directly...
if __name__ == "__main__":
    raise Exception("Not implemented")



"""Module to make it easy to start a validating session."""

import argparse
import logging
import shutil
import sys
import traceback
from pathlib import Path

from orthoseg.helpers import config_helper as conf, email_helper
from orthoseg.lib import prepare_traindatasets as prep
from orthoseg.util import log_util

# Get a logger...
logger = logging.getLogger(__name__)


def _validate_args(args) -> argparse.Namespace:
    # Interprete arguments
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group("Required arguments")
    required.add_argument(
        "-c", "--config", type=str, required=True, help="The config file to use"
    )

    # Optional arguments
    optional = parser.add_argument_group("Optional arguments")
    # Add back help
    optional.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit",
    )
    optional.add_argument(
        "config_overrules",
        nargs="*",
        help=(
            "Supply any number of config overrules like this: "
            "<section>.<parameter>=<value>"
        ),
    )

    return parser.parse_args(args)


def validate(config_path: Path, config_overrules: list[str] = []):
    """Run a validating session for the config specified.

    Args:
        config_path (Path): Path to the config file to use.
        config_overrules (list[str], optional): list of config options that will
            overrule other ways to supply configuration. They should be specified in the
            form of "<section>.<parameter>=<value>". Defaults to [].
    """
    # Init
    # Load the config and save in a bunch of global variables so it
    # is accessible everywhere
    conf.read_orthoseg_config(config_path, overrules=config_overrules)

    # Init logging
    log_util.clean_log_dir(
        log_dir=conf.dirs.getpath("log_dir"),
        nb_logfiles_tokeep=conf.logging_conf.getint("nb_logfiles_tokeep"),
    )
    global logger
    logger = log_util.main_log_init(conf.dirs.getpath("log_dir"), __name__)

    # Log start
    logger.info(f"Start validate for config {config_path.stem}")
    logger.debug(f"Config used: \n{conf.pformat_config()}")

    try:
        # Create the output dir's if they don't exist yet...
        for dir in [
            conf.dirs.getpath("project_dir"),
            conf.dirs.getpath("training_dir"),
        ]:
            if dir and not dir.exists():
                dir.mkdir()

        train_label_infos = conf.get_train_label_infos()
        classes = conf.determine_classes()

        # Now create the train datasets (train, validation, test)
        logger.info("Prepare train, validation and test data")
        training_dir, traindata_id = prep.prepare_traindatasets(
            label_infos=train_label_infos,
            classes=classes,
            image_layers=conf.image_layers,
            training_dir=conf.dirs.getpath("training_dir"),
            labelname_column=conf.train.get("labelname_column"),
            image_pixel_x_size=conf.train.getfloat("image_pixel_x_size"),
            image_pixel_y_size=conf.train.getfloat("image_pixel_y_size"),
            image_pixel_width=conf.train.getint("image_pixel_width"),
            image_pixel_height=conf.train.getint("image_pixel_height"),
            ssl_verify=conf.general["ssl_verify"],
            only_validate=True,
        )

        # Send mail that validate was successful
        email_helper.sendmail(f"Validate for config {config_path.stem} was successful")
        logger.info(
            f"Traindata dir to use is {training_dir}, with traindata_id: {traindata_id}"
        )
    except Exception as ex:
        message = f"ERROR while running validate for task {config_path.stem}"
        logger.exception(message)
        if isinstance(ex, prep.ValidationError):
            message_body = f"Validation error: {ex.to_html()}"
        else:
            message_body = f"Exception: {ex}<br/><br/>{traceback.format_exc()}"
        email_helper.sendmail(subject=message, body=message_body)
        raise Exception(message) from ex
    finally:
        if conf.tmp_dir is not None:
            shutil.rmtree(conf.tmp_dir, ignore_errors=True)


def main():
    """Run validate."""
    try:
        # Interprete arguments
        args = _validate_args(sys.argv[1:])

        # Run!
        validate(config_path=Path(args.config), config_overrules=args.config_overrules)
    except Exception as ex:
        logger.exception(f"Error: {ex}")
        raise


# If the script is ran directly...
if __name__ == "__main__":
    main()



"""Module with specific helper functions to get info for gisdata files."""

import logging
from pathlib import Path
from typing import Optional

logger = logging.getLogger(__name__)

# ---------------------------------------------------------
# Functions to get base paths of locations where GIS data
# can be stored
# ---------------------------------------------------------


def aidetection_dir() -> Path:
    """AI detection path.

    Returns:
        Path: AI detection path
    """
    return Path("X:/Monitoring/orthoseg")


# ---------------------------------------------------------
# Functions to get info about AI detection files
# ---------------------------------------------------------


class AiDetectionInfo:
    """Information about an ai detection result."""

    def __init__(
        self,
        path: Path,
        subject: str,
        traindata_version: int,
        image_layer: str,
        image_layer_year: Optional[int],
        postprocessing: str,
    ):
        """Init.

        Args:
            path (Path): the path to the file
            subject (str): the detection/segment subject
            traindata_version (int): the version of the data used to train the model
            image_layer (str): the image layer that the detection is based on
            image_layer_year (int): the year the images in the image_layer were taken
            postprocessing (str): the postprocessing done
        """
        self.path = path
        self.subject = subject
        self.traindata_version = traindata_version
        self.image_layer = image_layer
        self.image_layer_year = image_layer_year
        self.postprocessing = postprocessing


def aidetection_info(path: Path) -> AiDetectionInfo:
    """Get the properties from an ai detection filepath.

    Args:
        path (Path): the filepath to the ai detection file

    Return:
        AiDetectionInfo: info about the detection.
    """
    # Extract the fields...
    try:
        param_values = path.stem.split("_")
        if len(param_values) < 4:
            logger.warning(
                "No valid path for an ai detection, split('_') should result in 4 "
                f"fields: {path}"
            )

        subject = param_values[0]
        model_info = param_values[1]
        model_info_values = model_info.split(".")
        traindata_version = int(model_info_values[0])
        """
        architecture_version = 0
        trainparams_version = 0
        if len(model_info_values) > 1:
            architecture_version = int(model_info_values[1])
            if len(model_info_values) > 2:
                trainparams_version = int(model_info_values[2])
        aimodel_epoch = int(param_values[2])
        """
        image_layer = param_values[3]
        image_layer_values = image_layer.split("-")
        try:
            image_layer_year = int(image_layer_values[1])
        except ValueError:
            image_layer_year = None
        if len(param_values) > 4:
            postprocessing = "_".join(param_values[4:])
        else:
            postprocessing = ""

    except Exception as ex:
        raise ValueError(f"Error in get_aidetection_info on {path}") from ex

    return AiDetectionInfo(
        path, subject, traindata_version, image_layer, image_layer_year, postprocessing
    )



"""Module with functions for post-processing prediction masks towards polygons."""

import logging
import math
import shutil
from pathlib import Path
from typing import Optional

import geofileops as gfo
import geopandas as gpd
import numpy as np
import pygeoops
import rasterio as rio
import rasterio.features as rio_features
import rasterio.transform as rio_transform
import shapely.geometry as sh_geom
import skimage.filters.rank
import tensorflow as tf
from skimage.morphology import rectangle

from orthoseg.helpers import vectorfile_helper
from orthoseg.util import vector_util

# Avoid having many info warnings about self intersections from shapely
logging.getLogger("shapely.geos").setLevel(logging.WARNING)

# Get a logger...
logger = logging.getLogger(__name__)

# -------------------------------------------------------------
# Postprocess to use on all vector outputs
# -------------------------------------------------------------


def postprocess_predictions(
    input_path: Path,
    output_path: Path,
    dissolve: bool,
    dissolve_tiles_path: Optional[Path] = None,
    reclassify_to_neighbour_query: Optional[str] = None,
    simplify_algorithm: Optional[str] = None,
    simplify_tolerance: float = 1,
    simplify_lookahead: int = 8,
    keep_original_file: bool = True,
    keep_intermediary_files: bool = True,
    nb_parallel: int = -1,
    force: bool = False,
) -> list[Path]:
    """Postprocesses the input prediction as specified.

    Args:
        input_path: path to the 'raw' prediction vector file.
        output_path: the base path where the output file(s) will be written to.
        keep_original_file: If True, the output file of the prediction step
            will be retained ofter postprocessing, otherwise it is removed.
        keep_intermediary_files: If True, intermediary postprocessing files are removed.
        dissolve (bool): True if a dissolve needs to be applied
        dissolve_tiles_path (PathLike, optional): Path to a geofile containing
            the tiles to be used for the dissolve. Defaults to None.
        reclassify_to_neighbour_query (str, optional): Defaults to None.
        simplify_algorithm (str, optional): Algorithm to use for simplification. If
            None, no simplification is applied. Defaults to None.
        simplify_tolerance (float): Tolerance to use for the simplification.
            Defaults to 1.
        simplify_lookahead (int): Lookahead to use for simplification. Default to 8.
        nb_parallel (int, optional): number of cpu's to use for postprocessing.
            Use all cpu's if it is -1. Defaults to -1.
        force: False to skip results that already exist, true to
               ignore existing results and overwrite them
    """
    # Init
    if not input_path.exists():
        raise Exception(f"input_path does not exist: {input_path}")

    # The return value is the list of paths created
    output_paths = []

    # Because the geo operations will be applied sequentially if applicable,
    # both the input path and output path will build on the result of the
    # previous operation.
    # Set the initial values to the ones passed in as parameters.
    curr_input_path = input_path
    curr_output_path = output_path

    # Dissolve the predictions if needed
    if dissolve:
        curr_output_path = (
            output_path.parent / f"{output_path.stem}_dissolve{output_path.suffix}"
        )

        # If the dissolved file doesn't exist yet, go for it...
        if not curr_output_path.exists():
            # If column classname present, group on it...
            layerinfo = gfo.get_layerinfo(input_path)
            if "classname" in layerinfo.columns:
                groupby_columns = ["classname"]
            else:
                groupby_columns = []

            # Now we can dissolve
            gfo.dissolve(
                input_path=input_path,
                tiles_path=dissolve_tiles_path,
                output_path=curr_output_path,
                groupby_columns=groupby_columns,
                explodecollections=True,
                nb_parallel=nb_parallel,
                force=force,
            )

            # Add/recalculate columns with area and nbcoords
            gfo.add_column(
                path=curr_output_path,
                name="area",
                type=gfo.DataType.REAL,
                expression="ST_Area(geom)",
                force_update=True,
            )
            gfo.add_column(
                path=curr_output_path,
                name="nbcoords",
                type=gfo.DataType.INTEGER,
                expression="ST_NPoints(geom)",
                force_update=True,
            )

        # The curr_output_path becomes the new current input path
        curr_input_path = curr_output_path
        output_paths.append(curr_output_path)

    if reclassify_to_neighbour_query is not None:
        curr_output_path = (
            output_path.parent / f"{output_path.stem}_reclass{output_path.suffix}"
        )
        vectorfile_helper.reclassify_neighbours(
            input_path=curr_input_path,
            reclassify_column="classname",
            query=reclassify_to_neighbour_query,
            output_path=curr_output_path,
        )
        curr_input_path = curr_output_path
        output_paths.append(curr_output_path)

    # If a simplify algorithm is specified, simplify!
    if simplify_algorithm is not None:
        curr_input_path = curr_output_path
        curr_output_path = (
            curr_output_path.parent
            / f"{curr_output_path.stem}_simpl{curr_output_path.suffix}"
        )

        # If the simplified file doesn't exist yet, go for it...
        if not curr_output_path.exists():
            # Simplify!
            gfo.simplify(
                input_path=curr_input_path,
                output_path=curr_output_path,
                algorithm=gfo.SimplifyAlgorithm(simplify_algorithm),
                tolerance=simplify_tolerance,
                lookahead=simplify_lookahead,
                nb_parallel=nb_parallel,
            )

            # Add/recalculate columns with area and nbcoords
            gfo.add_column(
                path=curr_output_path,
                name="area",
                type=gfo.DataType.REAL,
                expression="ST_Area(geom)",
                force_update=True,
            )
            gfo.add_column(
                path=curr_output_path,
                name="nbcoords",
                type=gfo.DataType.INTEGER,
                expression="ST_NPoints(geom)",
                force_update=True,
            )

        curr_input_path = curr_output_path
        output_paths.append(curr_output_path)

    # If postprocessing steps are defined, the output of the prediction step
    # (input_path) is renamed to ..._orig.gpkg
    if dissolve or reclassify_to_neighbour_query or simplify_algorithm:
        original_file = input_path.parent / f"{input_path.stem}_orig.gpkg"
        input_path.rename(original_file)
        shutil.copy(src=curr_output_path, dst=input_path)

    # Cleanup original file
    if not keep_original_file:
        original_file.unlink()

    # Cleanup intermediary files
    if not keep_intermediary_files:
        for file in output_paths:
            file.unlink()

    return output_paths


def read_prediction_file(
    filepath: Path, border_pixels_to_ignore: int = 0
) -> Optional[gpd.GeoDataFrame]:
    """Read the prediction file specified.

    Args:
        filepath (Path): path to the prediction file.
        border_pixels_to_ignore (int, optional): border pixels the should be ignored.
            Defaults to 0.

    Returns:
        Optional[gpd.GeoDataFrame]: the vectorized and cleaned prediction.
    """
    ext_lower = filepath.suffix.lower()
    if ext_lower == ".geojson":
        return gfo.read_file(filepath)
    elif ext_lower == ".tif":
        return polygonize_pred_from_file(filepath, border_pixels_to_ignore)
    else:
        raise ValueError(f"Unsupported extension: {ext_lower}")


def to_binary_uint8(in_arr: np.ndarray, thresshold_ok: int = 128) -> np.ndarray:
    """Convert input array to binary UINT8.

    Args:
        in_arr (np.ndarray): input array
        thresshold_ok (int, optional): thresshold to use. Defaults to 128.

    Returns:
        np.ndarray: result
    """
    # Check input parameters
    if in_arr.dtype != np.uint8:
        raise ValueError(f"Input should be dtype = uint8, not: {in_arr.dtype}")

    # First copy to new numpy array, otherwise input array is changed
    out_arr = np.copy(in_arr)
    out_arr[out_arr >= thresshold_ok] = 255
    out_arr[out_arr < thresshold_ok] = 0

    return out_arr


def postprocess_for_evaluation(
    image_filepath: Path,
    image_crs: str,
    image_transform,
    image_pred_filepath: Path,
    image_pred_uint8_cleaned_bin: np.ndarray,
    class_id: int,
    class_name: str,
    nb_classes: int,
    output_dir: Path,
    output_suffix: Optional[str] = None,
    input_image_dir: Optional[Path] = None,
    input_mask_dir: Optional[Path] = None,
    border_pixels_to_ignore: int = 0,
    force: bool = False,
):
    """This function postprocesses a prediction for manual evaluation.

    To make it easy to evaluate visually if the result is OK by creating images of the
    different stages of the prediction logic by creating the following output:
        - the input image
        - the mask image as digitized in the train files (if available)
        - the "raw" prediction image
        - the "raw" polygonized prediction, as an image
        - the simplified polygonized prediction, as an image

    The filenames start with a prefix:
        - if a mask is available, the % overlap between the result and the mask
        - if no mask is available, the % of pixels that is white

    Args:
        image_filepath (Path): _description_
        image_crs (str): _description_
        image_transform (_type_): _description_
        image_pred_filepath (Path): _description_
        image_pred_uint8_cleaned_bin (np.ndarray): _description_
        class_id (int): _description_
        class_name (str): _description_
        nb_classes (int): _description_
        output_dir (Path): _description_
        output_suffix (Optional[str], optional): _description_. Defaults to None.
        input_image_dir (Optional[Path], optional): _description_. Defaults to None.
        input_mask_dir (Optional[Path], optional): _description_. Defaults to None.
        border_pixels_to_ignore (int, optional): _description_. Defaults to 0.
        force (bool, optional): _description_. Defaults to False.

    Raises:
        ValueError: _description_
        Exception: _description_
        Exception: _description_

    Returns:
        _type_: _description_
    """
    logger.debug(f"Start postprocess for {image_pred_filepath}")
    all_black = False
    try:
        # If the image wasn't saved, it must have been all black
        if image_pred_filepath is None:
            all_black = True

        # Make sure the output dir exists...
        output_dir.mkdir(parents=True, exist_ok=True)

        # Determine the prefix to use for the output filenames
        pred_prefix_str = ""
        """
        def jaccard_similarity(im1: np.ndarray, im2: np.ndarray):
            if im1.shape != im2.shape:
                message = f"input shapes mismatch: im1: {im1.shape}, im2: {im2.shape}"
                logger.critical(message)
                raise ValueError(message)

            intersection = np.logical_and(im1, im2)
            union = np.logical_or(im1, im2)

            sum_union = float(union.sum())
            if sum_union == 0.0:
                # If 0 positive pixels in union: perfect prediction, so 1
                return 1
            else:
                sum_intersect = intersection.sum()
                return sum_intersect/sum_union
        """

        # If there is a mask dir specified... use the groundtruth mask
        if input_mask_dir is not None and input_mask_dir.exists():
            # Read mask file and get all needed info from it...
            mask_filepath = Path(
                str(image_filepath).replace(str(input_image_dir), str(input_mask_dir))
            )

            # Check if this file exists, if not, look for similar files
            if not mask_filepath.exists():
                files = list(mask_filepath.parent.glob(mask_filepath.stem + "*"))
                if len(files) == 1:
                    mask_filepath = files[0]
                else:
                    message = (
                        f"Error finding mask file with {mask_filepath.stem + '*'}: "
                        f"{len(files)} mask(s) found"
                    )
                    logger.error(message)
                    raise Exception(message)

            with rio.open(mask_filepath) as mask_ds:
                # Read pixels
                mask_arr = mask_ds.read(1)

            # Make the pixels at the borders of the mask black so they are
            # ignored in the comparison
            if border_pixels_to_ignore and border_pixels_to_ignore > 0:
                mask_arr[0:border_pixels_to_ignore, :] = 0  # Left border
                mask_arr[-border_pixels_to_ignore:, :] = 0  # Right border
                mask_arr[:, 0:border_pixels_to_ignore] = 0  # Top border
                mask_arr[:, -border_pixels_to_ignore:] = 0  # Bottom border

            # If there is more than 1 class, extract the seperate masks
            # per class with one-hot encoding
            if nb_classes > 1:
                mask_categorical_arr = tf.keras.utils.to_categorical(
                    mask_arr, nb_classes, dtype=rio.uint8
                )
                mask_arr = (mask_categorical_arr[:, :, class_id]) * 255

            # similarity = jaccard_similarity(mask_arr, image_pred)
            # Use accuracy as similarity... is more practical than jaccard
            similarity = (
                np.array(np.equal(mask_arr, image_pred_uint8_cleaned_bin)).sum()
                / image_pred_uint8_cleaned_bin.size
            )
            pred_prefix_str = f"{similarity:0.3f}_"

            # Write mask
            mask_copy_dest_filepath = (
                output_dir
                / f"{pred_prefix_str}{image_filepath.stem}_{class_name}_mask.tif"
            )
            # if not mask_copy_dest_filepath.exists():
            with rio.open(
                mask_copy_dest_filepath,
                "w",
                driver="GTiff",
                compress="lzw",
                height=mask_arr.shape[1],
                width=mask_arr.shape[0],
                count=1,
                dtype=rio.uint8,
                crs=image_crs,
                transform=image_transform,
            ) as dst:
                dst.write(mask_arr, 1)

        else:
            # If all_black, no need to calculate again
            if all_black:
                pct_black = 1
            else:
                # Calculate percentage black pixels
                pct_black = 1 - (
                    (image_pred_uint8_cleaned_bin.sum() / 250)
                    / image_pred_uint8_cleaned_bin.size
                )

            # If the result after segmentation is all black, set all_black
            if pct_black == 1:
                # Force the prefix to be high so it is clear they are entirely black
                pred_prefix_str = "1.001_"
                all_black = True
            else:
                pred_prefix_str = f"{pct_black:0.3f}_"

            # If there are few white pixels, don't save it,
            # because we are in evaluetion mode anyway...
            # if similarity >= 0.95:
            # continue

        # Copy the input image if it doesn't exist yet in output path
        output_basefilepath = (
            output_dir / f"{pred_prefix_str}{image_filepath.stem}{output_suffix}"
        )
        image_dest_filepath = Path(str(output_basefilepath) + image_filepath.suffix)
        if not image_dest_filepath.exists():
            shutil.copyfile(image_filepath, image_dest_filepath)

        # Rename the prediction file so it also contains the prefix,...
        if image_pred_filepath is not None:
            image_dest_filepath = Path(
                f"{output_basefilepath!s}_pred{image_pred_filepath.suffix}"
            )
            if not image_dest_filepath.exists():
                shutil.move(str(image_pred_filepath), image_dest_filepath)

        # If all_black, we are ready now
        if all_black:
            logger.debug("All black prediction, no use proceding")
            return

        # Write a cleaned-up version for evaluation as well
        # polygonize_pred_for_evaluation(
        #        image_pred_uint8_bin=image_pred_uint8_cleaned_bin,
        #        image_crs=image_crs,
        #        image_transform=image_transform,
        #        output_basefilepath=output_basefilepath)

    except Exception as ex:
        message = (
            f"Error postprocessing prediction for {image_filepath}:\n"
            f"    file: {image_pred_filepath}!!!"
        )
        raise Exception(message) from ex


def polygonize_pred_for_evaluation(
    image_pred_uint8_bin, image_crs: str, image_transform, output_basefilepath: Path
):
    """Polygonize a prediction in a way it is easy to evaluate manually/visually.

    Args:
        image_pred_uint8_bin (_type_): _description_
        image_crs (str): _description_
        image_transform (_type_): _description_
        output_basefilepath (Path): _description_
    """
    # Polygonize result
    try:
        # Returns a list of tupples with (geometry, value)
        polygonized_records = list(
            rio_features.shapes(
                image_pred_uint8_bin,
                mask=image_pred_uint8_bin,
                transform=image_transform,
            )
        )

        # If nothing found, we can return
        if len(polygonized_records) == 0:
            logger.debug("This prediction didn't result in any polygons")
            return

        # Convert shapes to geopandas geodataframe
        geoms = []
        for geom, _ in polygonized_records:
            geoms.append(sh_geom.shape(geom))
        geoms_gdf = gpd.GeoDataFrame(geoms, columns=["geometry"])
        geoms_gdf.crs = image_crs

        image_shape = image_pred_uint8_bin.shape
        image_width = image_shape[0]
        image_height = image_shape[1]

        # For easier evaluation, write the cleaned version as raster
        # Write the standard cleaned output to file
        logger.debug("Save binary prediction")
        image_pred_cleaned_filepath = Path(f"{output_basefilepath!s}_pred_bin.tif")
        with rio.open(
            image_pred_cleaned_filepath,
            "w",
            driver="GTiff",
            compress="lzw",
            height=image_height,
            width=image_width,
            count=1,
            dtype=rio.uint8,
            crs=image_crs,
            transform=image_transform,
        ) as dst:
            dst.write(image_pred_uint8_bin, 1)

        # If the input image contained a transform, also create an image
        # based on the simplified vectors
        if image_transform[0] != 0 and len(geoms) > 0:
            # Simplify geoms
            geoms_simpl = []
            for geom in geoms:
                # The simplify of shapely uses the deuter-pecker algo
                # preserve_topology is slower bu makes sure no polygons are removed
                geom_simpl = geom.simplify(0.5, preserve_topology=True)
                if not geom_simpl.is_empty:
                    geoms_simpl.append(geom_simpl)

            # Write simplified wkt result to raster for comparing.
            if len(geoms_simpl) > 0:
                # TODO: doesn't support multiple classes
                logger.debug("Before writing simpl rasterized file")
                image_pred_simpl_filepath = (
                    f"{output_basefilepath!s}_pred_cleaned_simpl.tif"
                )
                with rio.open(
                    image_pred_simpl_filepath,
                    "w",
                    driver="GTiff",
                    compress="lzw",
                    height=image_height,
                    width=image_width,
                    count=1,
                    dtype=rio.uint8,
                    crs=image_crs,
                    transform=image_transform,
                ) as dst:
                    # create a generator of geom, value pairs to use in rasterizing
                    logger.debug("Before rasterize")
                    burned = rio_features.rasterize(
                        shapes=geoms_simpl,
                        out_shape=(image_height, image_width),
                        fill=0,
                        default_value=255,
                        dtype=rio.uint8,
                        transform=image_transform,
                    )
                    dst.write(burned, 1)

    except Exception as ex:
        message = f"Exception while polygonizing to file {output_basefilepath}!"
        raise RuntimeError(message) from ex


def polygonize_pred_from_file(
    image_pred_filepath: Path,
    border_pixels_to_ignore: int = 0,
    save_to_file: bool = False,
) -> Optional[gpd.GeoDataFrame]:
    """Polygonize a prediction from a file.

    Args:
        image_pred_filepath (Path): path to the file to be read.
        border_pixels_to_ignore (int, optional): number of pixels around the input image
            to ignore. Defaults to 0.
        save_to_file (bool, optional): If True, save te result to a file.
            Defaults to False.

    Returns:
        Optional[gpd.GeoDataFrame]: _description_
    """
    try:
        with rio.open(image_pred_filepath) as image_ds:
            # Read geo info
            image_crs = image_ds.profile["crs"]
            image_transform = image_ds.transform

            # Read pixels and change from (channels, width, height) to
            # (width, height, channels) and normalize to values between 0 and 1
            image_data = image_ds.read()

        # Create binary version
        # image_data = rio_plot.reshape_as_image(image_data)
        image_pred_uint8_bin = to_binary_uint8(image_data, 125)

        output_basefilepath = None
        if save_to_file is True:
            output_basefilepath = image_pred_filepath.parent / image_pred_filepath.stem
        result_gdf = polygonize_pred(
            image_pred_uint8_bin=image_pred_uint8_bin,
            image_crs=image_crs,
            image_transform=image_transform,
            output_basefilepath=output_basefilepath,
        )

        if result_gdf is None:
            logger.warning(
                f"Prediction didn't result in any polygons: {image_pred_filepath}"
            )

        return result_gdf

    except Exception as ex:
        raise RuntimeError(
            f"Error in polygonize_pred_from_file on {image_pred_filepath}"
        ) from ex


def polygonize_pred_multiclass_to_file(
    image_pred_arr: np.ndarray,
    image_crs: str,
    image_transform,
    classes: list,
    output_vector_path: Path,
    min_probability: float = 0.5,
    postprocess: dict = {},
    border_pixels_to_ignore: int = 0,
    create_spatial_index: bool = True,
) -> dict:
    """Polygonize a multiclass prediction to a file.

    Args:
        image_pred_arr (np.ndarray): _description_
        image_crs (str): _description_
        image_transform (_type_): _description_
        classes (list): _description_
        output_vector_path (Path): _description_
        min_probability (float, optional): _description_. Defaults to 0.5.
        postprocess (dict, optional): _description_. Defaults to {}.
        border_pixels_to_ignore (int, optional): _description_. Defaults to 0.
        create_spatial_index (bool, optional): _description_. Defaults to True.

    Returns:
        dict: _description_
    """
    # Polygonize the result...
    result_gdf = polygonize_pred_multiclass(
        image_pred_uint8=image_pred_arr,
        image_crs=image_crs,
        image_transform=image_transform,
        classes=classes,
        min_probability=min_probability,
        postprocess=postprocess,
        border_pixels_to_ignore=border_pixels_to_ignore,
    )

    # If there were polygons, save them...
    if result_gdf is not None:
        gfo.to_file(
            result_gdf,
            output_vector_path,
            append=True,
            index=False,
            force_multitype=True,
            create_spatial_index=create_spatial_index,
        )
        return {"nb_features_witten": len(result_gdf), "columns": result_gdf.columns}
    else:
        return {"nb_features_witten": None}


def polygonize_pred_multiclass(
    image_pred_uint8: np.ndarray,
    image_crs: str,
    image_transform,
    classes: list,
    min_probability: float = 0.5,
    postprocess: dict = {},
    border_pixels_to_ignore: int = 0,
) -> Optional[gpd.GeoDataFrame]:
    """Polygonize a multiclass prediction.

    Args:
        image_pred_uint8 (np.ndarray): _description_
        image_crs (str): _description_
        image_transform (_type_): _description_
        classes (list): _description_
        min_probability (float, optional): _description_. Defaults to 0.5.
        postprocess (dict, optional): _description_. Defaults to {}.
        border_pixels_to_ignore (int, optional): _description_. Defaults to 0.

    Returns:
        Optional[gpd.GeoDataFrame]: _description_
    """
    # Init
    """
    for channel_id in range(0, nb_channels):
        image_pred_curr_arr = image_pred_arr[:,:,channel_id]

        # Clean prediction
        image_pred_uint8_cleaned_curr = clean_prediction(
                image_pred_arr=image_pred_curr_arr,
                border_pixels_to_ignore=border_pixels_to_ignore)
    """

    # Reverse the one-hot decoding so each class has it's own number in the array,
    # but ignore prediction probability < min_probability
    image_pred_uint8[image_pred_uint8 < math.floor(255 * min_probability)] = 0
    image_pred_decoded_arr = np.argmax(image_pred_uint8, axis=2).astype(np.uint8)

    # Make the pixels at the borders of the prediction black so they are ignored
    if border_pixels_to_ignore and border_pixels_to_ignore > 0:
        image_pred_decoded_arr[0:border_pixels_to_ignore, :] = 0  # Left border
        image_pred_decoded_arr[-border_pixels_to_ignore:, :] = 0  # Right border
        image_pred_decoded_arr[:, 0:border_pixels_to_ignore] = 0  # Top border
        image_pred_decoded_arr[:, -border_pixels_to_ignore:] = 0  # Bottom border

    # Postprocessing on the raster output
    if len(postprocess) > 0:
        # If fill_gaps_modal_size is asked...
        if (
            "filter_background_modal_size" in postprocess
            and postprocess["filter_background_modal_size"] is not None
            and postprocess["filter_background_modal_size"] > 0
        ):
            filter_background_modal_size = postprocess["filter_background_modal_size"]
            image_pred_decoded_modal_arr = skimage.filters.rank.modal(
                image_pred_decoded_arr,
                rectangle(filter_background_modal_size, filter_background_modal_size),
            )
            np.copyto(
                image_pred_decoded_arr,
                image_pred_decoded_modal_arr,
                where=image_pred_decoded_arr == 0,
            )

    # Polygonize
    # If a reclassify query is specified don't mask so the query is also applied to
    # the background
    mask_background = True
    if (
        len(postprocess) > 0
        and postprocess.get("reclassify_to_neighbour_query", None) is not None
    ):
        mask_background = False

    result_gdf = polygonize_pred(
        image_pred_uint8_bin=image_pred_decoded_arr,
        image_crs=image_crs,
        image_transform=image_transform,
        mask_background=mask_background,
        classnames=classes,
    )
    if result_gdf is None:
        return None

    # Calculate the bounds of the image in projected coordinates
    image_shape = image_pred_decoded_arr.shape
    image_width = image_shape[0]
    image_height = image_shape[1]
    image_bounds = rio_transform.array_bounds(
        image_height, image_width, image_transform
    )
    x_pixsize = get_pixelsize_x(image_transform)
    y_pixsize = get_pixelsize_y(image_transform)
    border_bounds = (
        image_bounds[0] + border_pixels_to_ignore * x_pixsize,
        image_bounds[1] + border_pixels_to_ignore * y_pixsize,
        image_bounds[2] - border_pixels_to_ignore * x_pixsize,
        image_bounds[3] - border_pixels_to_ignore * y_pixsize,
    )

    # Postprocessing on the vectorized result
    if len(postprocess) > 0:
        # If a reclassify query is specified
        reclassify_query = postprocess.get("reclassify_to_neighbour_query", None)
        if reclassify_query is not None:
            result_gdf = vector_util.reclassify_neighbours(
                result_gdf,
                reclassify_column="classname",
                query=reclassify_query,
                border_bounds=border_bounds,
                class_background=classes[0],
            )

        # If a simplify is asked...
        simplify = postprocess.get("simplify", None)
        if simplify is not None:
            # Define the bounds of the image as linestring, so points on this
            # border are preserved during the simplify
            border_polygon = sh_geom.box(*border_bounds)
            assert border_polygon.exterior is not None
            border_lines = sh_geom.LineString(border_polygon.exterior.coords)

            # Determine of topological or normal simplify needs to be used
            simplify_topological = simplify["simplify_topological"]
            if simplify_topological is None:
                simplify_topological = True if len(classes) > 2 else False

            assert isinstance(result_gdf.geometry, gpd.GeoSeries)
            result_gdf.geometry = pygeoops.simplify(
                geometry=result_gdf.geometry,
                algorithm=simplify["simplify_algorithm"],
                tolerance=simplify["simplify_tolerance"],
                lookahead=simplify["simplify_lookahead"],
                preserve_common_boundaries=simplify_topological,
                keep_points_on=border_lines,
            )

            # Remove geom rows that became empty after simplify + explode
            assert result_gdf.geometry is not None
            result_gdf = result_gdf[~result_gdf.geometry.is_empty]
            result_gdf = result_gdf[~result_gdf.geometry.isna()]
            if len(result_gdf) == 0:
                return None
            result_gdf = result_gdf.explode(ignore_index=True)

    assert isinstance(result_gdf, gpd.GeoDataFrame)
    return result_gdf


def polygonize_pred(
    image_pred_uint8_bin,
    image_crs: str,
    image_transform,
    mask_background: bool = True,
    classnames: Optional[list[str]] = None,
    output_basefilepath: Optional[Path] = None,
) -> Optional[gpd.GeoDataFrame]:
    """Polygonize a prediction.

    Args:
        image_pred_uint8_bin (_type_): _description_
        image_crs (str): _description_
        image_transform (_type_): _description_
        mask_background (bool, optional): _description_. Defaults to True.
        classnames (Optional[list[str]], optional): _description_. Defaults to None.
        output_basefilepath (Optional[Path], optional): _description_. Defaults to None.

    Raises:
        Exception: _description_

    Returns:
        Optional[gpd.GeoDataFrame]: _description_
    """
    # Polygonize result
    try:
        # Returns a list of tupples with (geometry, value)
        mask = None
        if mask_background:
            mask = image_pred_uint8_bin
        polygonized_records = list(
            rio_features.shapes(
                image_pred_uint8_bin,
                mask=mask,
                transform=image_transform,
            )
        )

        # If nothing found, we can return
        if len(polygonized_records) == 0:
            return None

        # Convert shapes to geopandas geodataframe
        data = [
            (sh_geom.shape(geom), int(value)) for geom, value in polygonized_records
        ]
        result_gdf = gpd.GeoDataFrame(
            data, columns=["geometry", "value"], crs=image_crs
        )

        # Add the classname if provided
        if classnames is not None:
            result_gdf["classname"] = [
                classnames[value] for _, value in result_gdf["value"].T.items()
            ]
            result_gdf = result_gdf.drop(columns=["value"])

        assert isinstance(result_gdf, gpd.GeoDataFrame)
        return result_gdf

    except Exception as ex:
        message = f"Exception while polygonizing to file {output_basefilepath}"
        raise Exception(message) from ex


def clean_and_save_prediction(
    input_image_filepath: Path,
    image_crs: str,
    image_transform: str,
    output_dir: Path,
    image_pred_arr: np.ndarray,
    classes: list,
    input_image_dir: Optional[Path] = None,
    input_mask_dir: Optional[Path] = None,
    border_pixels_to_ignore: int = 0,
    min_probability: float = 0.5,
    evaluate_mode: bool = False,
    force: bool = False,
) -> bool:
    """Clean the prediction and save it.

    Args:
        input_image_filepath (Path): _description_
        image_crs (str): _description_
        image_transform (str): _description_
        output_dir (Path): _description_
        image_pred_arr (np.ndarray): _description_
        classes (list): _description_
        input_image_dir (Optional[Path], optional): _description_. Defaults to None.
        input_mask_dir (Optional[Path], optional): _description_. Defaults to None.
        border_pixels_to_ignore (int, optional): _description_. Defaults to 0.
        min_probability (float, optional): _description_. Defaults to 0.5.
        evaluate_mode (bool, optional): _description_. Defaults to False.
        force (bool, optional): _description_. Defaults to False.

    Raises:
        Exception: _description_

    Returns:
        bool: _description_
    """
    # If nb. channels in prediction > 1, skip the first as it is the background
    image_pred_shape = image_pred_arr.shape
    nb_channels = image_pred_shape[2]
    if nb_channels > 1:
        channel_start = 1
    else:
        channel_start = 0

    for channel_id in range(channel_start, nb_channels):
        image_pred_curr_arr = image_pred_arr[:, :, channel_id]

        # Clean prediction
        image_pred_uint8_cleaned_curr = clean_prediction(
            image_pred_arr=image_pred_curr_arr,
            border_pixels_to_ignore=border_pixels_to_ignore,
        )

        # If the cleaned result contains useful values or in evaluate mode... save
        if (
            min_probability == 0
            or np.any(
                image_pred_uint8_cleaned_curr >= math.floor(min_probability * 255)
            )
            or evaluate_mode is True
        ):
            # Find the class name in the classes list
            class_name = None
            for class_id, (classname) in enumerate(classes):
                if class_id == channel_id:
                    class_name = classname
                    break
            if class_name is None:
                raise Exception(f"No classname found for channel_id {channel_id}")

            # Now save prediction
            output_suffix = f"_{class_name}"
            image_pred_filepath = save_prediction_uint8(
                image_filepath=input_image_filepath,
                image_pred_uint8_cleaned=image_pred_uint8_cleaned_curr,
                image_crs=image_crs,
                image_transform=image_transform,
                output_dir=output_dir,
                output_suffix=output_suffix,
                force=force,
            )

            # Postprocess for evaluation
            if evaluate_mode is True:
                # Create binary version and postprocess
                image_pred_uint8_cleaned_bin = to_binary_uint8(
                    image_pred_uint8_cleaned_curr, 125
                )
                postprocess_for_evaluation(
                    image_filepath=input_image_filepath,
                    image_crs=image_crs,
                    image_transform=image_transform,
                    image_pred_filepath=image_pred_filepath,
                    image_pred_uint8_cleaned_bin=image_pred_uint8_cleaned_bin,
                    output_dir=output_dir,
                    output_suffix=output_suffix,
                    input_image_dir=input_image_dir,
                    input_mask_dir=input_mask_dir,
                    class_id=channel_id,
                    class_name=class_name,
                    nb_classes=nb_channels,
                    border_pixels_to_ignore=border_pixels_to_ignore,
                    force=force,
                )

    return True


def clean_prediction(
    image_pred_arr: np.ndarray,
    border_pixels_to_ignore: int = 0,
    output_color_depth: str = "binary",
) -> np.ndarray:
    """Cleans a prediction result and returns a cleaned, uint8 array.

    Args:
        image_pred_arr (np.array): The prediction as returned by keras.
        border_pixels_to_ignore (int, optional): Border pixels to ignore. Defaults to 0.
        output_color_depth (str, optional): Color depth desired. Defaults to '2'.
            * binary: 0 or 255
            * full: 256 different values

    Returns:
        np.array: The cleaned result.
    """
    # Input should be float32
    if image_pred_arr.dtype not in [np.float32, np.uint8]:
        raise Exception(
            f"image prediction is in an unsupported type: {image_pred_arr.dtype}"
        )
    if output_color_depth not in ["binary", "full"]:
        raise Exception(f"Unsupported output_color_depth: {output_color_depth}")

    # Reshape from 3 to 2 dims if necessary (width, height, nb_channels).
    # Check the number of channels of the output prediction
    image_pred_shape = image_pred_arr.shape
    if len(image_pred_shape) > 2:
        n_channels = image_pred_shape[2]
        if n_channels > 1:
            raise Exception("Invalid input, should be one channel!")
        # Reshape array from 3 dims (width, height, nb_channels) to 2.
        image_pred_uint8 = np.reshape(
            image_pred_arr, (image_pred_shape[0], image_pred_shape[1])
        )

    # Convert to uint8 if necessary
    if image_pred_arr.dtype == np.float32:
        image_pred_uint8 = np.array((image_pred_arr * 255), dtype=np.uint8)
    else:
        image_pred_uint8 = image_pred_arr

    # Convert to binary if needed
    if output_color_depth == "binary":
        image_pred_uint8[image_pred_uint8 >= 127] = 255
        image_pred_uint8[image_pred_uint8 < 127] = 0

    # Make the pixels at the borders of the prediction black so they are ignored
    image_pred_uint8_cropped = image_pred_uint8
    if border_pixels_to_ignore and border_pixels_to_ignore > 0:
        image_pred_uint8_cropped[0:border_pixels_to_ignore, :] = 0  # Left border
        image_pred_uint8_cropped[-border_pixels_to_ignore:, :] = 0  # Right border
        image_pred_uint8_cropped[:, 0:border_pixels_to_ignore] = 0  # Top border
        image_pred_uint8_cropped[:, -border_pixels_to_ignore:] = 0  # Bottom border

    return image_pred_uint8_cropped


def save_prediction_uint8(
    image_filepath: Path,
    image_pred_uint8_cleaned: np.ndarray,
    image_crs: str,
    image_transform: str,
    output_dir: Path,
    output_suffix: str = "",
    border_pixels_to_ignore: Optional[int] = None,
    force: bool = False,
) -> Path:
    """Save the prediction as UINT8.

    Args:
        image_filepath (Path): _description_
        image_pred_uint8_cleaned (np.ndarray): _description_
        image_crs (str): _description_
        image_transform (str): _description_
        output_dir (Path): _description_
        output_suffix (str, optional): _description_. Defaults to "".
        border_pixels_to_ignore (Optional[int], optional): _description_.
            Defaults to None.
        force (bool, optional): _description_. Defaults to False.

    Raises:
        Exception: _description_

    Returns:
        Path: _description_
    """
    # Init
    # If no decent transform metadata, stop!
    if image_transform is None or image_transform[0] == 0:
        message = f"No transform found for {image_filepath}: {image_transform}"
        logger.error(message)
        raise Exception(message)

    # Make sure the output dir exists...
    if not output_dir.exists():
        output_dir.mkdir()

    # Write prediction to file
    output_filepath = output_dir / f"{image_filepath.stem}{output_suffix}_pred.tif"
    logger.debug("Save +- original prediction")
    image_shape = image_pred_uint8_cleaned.shape
    image_width = image_shape[0]
    image_height = image_shape[1]
    with rio.open(
        str(output_filepath),
        "w",
        driver="GTiff",
        tiled="no",
        compress="lzw",
        predictor=2,
        num_threads=4,
        height=image_height,
        width=image_width,
        count=1,
        dtype=rio.uint8,
        crs=image_crs,
        transform=image_transform,
    ) as dst:
        dst.write(image_pred_uint8_cleaned, 1)

    return output_filepath


# -------------------------------------------------------------
# Helpers for working with Affine objects...
# -------------------------------------------------------------


def get_pixelsize_x(transform) -> float:
    """Get the x pixel size from the transform.

    Args:
        transform (_type_): input transform

    Returns:
        float: the x pixel size.
    """
    return transform[0]


def get_pixelsize_y(transform):
    """Get the y pixel size from the transform.

    Args:
        transform (_type_): input transform

    Returns:
        float: the y pixel size.
    """
    return -transform[4]



from contextlib import AbstractContextManager, nullcontext

import pytest

from orthoseg.lib import predicter


@pytest.mark.parametrize(
    "no_images_ok, exp_error",
    [
        (False, True),
        (True, False),
    ],
)
def test_predict_dir_input_image_dir_empty(
    tmp_path,
    no_images_ok: bool,
    exp_error: bool,
):
    input_image_dir = tmp_path / "input"
    input_image_dir.mkdir()
    output_image_dir = tmp_path / "output"
    output_image_dir.mkdir()

    handler: AbstractContextManager
    if exp_error:
        handler = pytest.raises(ValueError)
    else:
        handler = nullcontext()

    with handler:
        predicter.predict_dir(
            model=None,  # type: ignore
            input_image_dir=input_image_dir,
            output_image_dir=output_image_dir,
            output_vector_path=None,
            classes=[],
            no_images_ok=no_images_ok,
        )



"""
Tests for functionalities in orthoseg.train.
"""

import pyproj
import pytest
import rasterio as rio

from orthoseg.util import ows_util
from tests import test_helper


def test_get_images_for_grid(tmp_path):
    # Init some stuff
    filelayer_path = (
        test_helper.sampleprojects_dir
        / "fields/input_raster"
        / "BEFL-TEST-s2_2023-05-01_2023-07-01_B08-B04-B03_min_byte.tif"
    )
    # File bounds: left: 484500, bottom: 5642970, right: 499500, top: 5651030

    crs = "epsg:32631"
    pixsize_x = 5
    pixsize_y = pixsize_x
    width_pix = 512
    height_pix = 256
    width_crs = width_pix * pixsize_x
    height_crs = height_pix * pixsize_y
    pixels_overlap = 64
    xmin = 484500
    ymin = 5642970
    image_gen_bbox = (xmin, ymin, xmin + width_crs * 2, ymin + height_crs * 2)

    # Test getting the images for this grid
    # -------------------------------------
    layersource_rgb = ows_util.FileLayerSource(path=filelayer_path, layernames=["S1"])
    ows_util.get_images_for_grid(
        layersources=[layersource_rgb],
        output_image_dir=tmp_path,
        crs=crs,
        image_gen_bbox=image_gen_bbox,
        image_crs_pixel_x_size=pixsize_x,
        image_crs_pixel_y_size=pixsize_y,
        image_pixel_width=width_pix,
        image_pixel_height=height_pix,
        nb_concurrent_calls=1,
        image_format=ows_util.FORMAT_GEOTIFF,
        pixels_overlap=pixels_overlap,
        ssl_verify=True,
    )

    assert tmp_path.exists()
    image_paths = list(tmp_path.glob("**/*.tif"))

    assert len(image_paths) == 9
    for image_path in image_paths:
        assert image_path.exists()
        assert image_path.stat().st_size > 50000, "Image should be larger than 50kB"
        with rio.open(image_path) as image_file:
            assert image_file.width == width_pix + 2 * pixels_overlap
            assert image_file.height == height_pix + 2 * pixels_overlap


@pytest.mark.parametrize(
    "crs_epsg, has_switched_axes",
    [
        [4326, False],
        [31370, False],
        [3059, True],
        [31468, True],
    ],
)
def test_has_switched_axes(crs_epsg: int, has_switched_axes: bool):
    assert ows_util._has_switched_axes(pyproj.CRS(crs_epsg)) is has_switched_axes


def test_getmap_to_file_filelayer(tmp_path):
    # Init some stuff
    filelayer_path = (
        test_helper.sampleprojects_dir
        / "fields/input_raster"
        / "BEFL-TEST-s2_2023-05-01_2023-07-01_B08-B04-B03_min_byte.tif"
    )
    # File bounds: left: 484500, bottom: 5642970, right: 499500, top: 5651030
    with rio.open(filelayer_path) as filelayer:
        file_bounds = filelayer.bounds

    crs = "epsg:32631"
    pixsize_x = 5
    pixsize_y = pixsize_x
    width_pix = 512
    height_pix = 256
    width_crs = width_pix * pixsize_x
    height_crs = height_pix * pixsize_y
    xmin = 484400
    ymin = 5642900
    bbox = (xmin, ymin, xmin + width_crs, ymin + height_crs)

    # Align box to pixel size + make sure width stays the asked number of pixels
    bbox = ows_util.align_bbox_to_grid(
        bbox,
        grid_xmin=file_bounds[0],
        grid_ymin=file_bounds[1],
        pixel_size_x=pixsize_x,
        pixel_size_y=pixsize_y,
    )
    bbox = (bbox[0], bbox[1], bbox[0] + width_crs, bbox[1] + height_crs)

    # Test getting a standard 3 band image from a file layer
    # ------------------------------------------------------
    layersource_rgb = ows_util.FileLayerSource(path=filelayer_path, layernames=["S1"])
    image_path = ows_util.getmap_to_file(
        layersources=layersource_rgb,
        output_dir=tmp_path,
        crs=crs,
        bbox=bbox,
        size=(width_pix, height_pix),
        image_pixels_ignore_border=0,
        transparent=False,
        layername_in_filename=True,
    )

    assert image_path is not None
    assert image_path.exists()
    assert image_path.stat().st_size > 50000, "Image should be larger than 50kB"
    with rio.open(image_path) as image_file:
        assert tuple(image_file.bounds) == bbox
        assert image_file.width == width_pix
        assert image_file.height == height_pix


def test_getmap_to_file_wmslayer(tmp_path):
    # Init some stuff
    projection = "epsg:31370"
    pixsize_x = 0.25
    pixsize_y = pixsize_x
    width_pix = 512
    height_pix = 256
    width_crs = width_pix * pixsize_x
    height_crs = height_pix * pixsize_y
    xmin = 160000
    ymin = 170000
    bbox = (xmin, ymin, xmin + width_crs, ymin + height_crs)

    # Test getting a standard 3 band image from a WMS layer
    # -----------------------------------------------------
    layersource_rgb = ows_util.WMSLayerSource(
        wms_server_url="https://geo.api.vlaanderen.be/omw/wms?",
        layernames=["OMWRGB20VL"],
        layerstyles=["default"],
    )
    image_path = ows_util.getmap_to_file(
        layersources=layersource_rgb,
        output_dir=tmp_path,
        crs=projection,
        bbox=bbox,
        size=(width_pix, height_pix),
        image_format=ows_util.FORMAT_PNG,
        image_pixels_ignore_border=0,
        transparent=False,
        layername_in_filename=True,
    )

    assert image_path is not None
    assert image_path.exists()
    assert image_path.stat().st_size > 50000, "Image should be larger than 50kB"
    with rio.open(image_path) as image_file:
        # assert tuple(image_file.bounds) == bbox
        assert image_file.width == width_pix
        assert image_file.height == height_pix

    # Test creating greyscale image
    # -----------------------------
    # If band -1 is specified, a greyscale version of the rgb image will be created
    layersource_dhm_ortho_grey = ows_util.WMSLayerSource(
        wms_server_url="https://geo.api.vlaanderen.be/ogw/wms?",
        layernames=["OGWRGB13_15VL"],
        bands=[-1],
    )
    image_path = ows_util.getmap_to_file(
        layersources=layersource_dhm_ortho_grey,
        output_dir=tmp_path,
        crs=projection,
        bbox=bbox,
        size=(width_pix, height_pix),
        image_format=ows_util.FORMAT_PNG,
        # image_format_save=ows_util.FORMAT_TIFF,
        image_pixels_ignore_border=0,
        transparent=False,
        layername_in_filename=True,
    )

    assert image_path is not None
    assert image_path.exists() is True
    with rio.open(image_path) as image_file:
        # assert tuple(image_file.bounds) == bbox
        assert image_file.width == width_pix
        assert image_file.height == height_pix

    # Test combining bands of 3 different WMS layers
    # ----------------------------------------------
    # Layer sources for for skyview and hillshade
    wms_server_url_dhm = "https://geo.api.vlaanderen.be/DHMV/wms?"
    layersource_dhm_skyview = ows_util.WMSLayerSource(
        wms_server_url=wms_server_url_dhm,
        layernames=["DHMV_II_SVF_25cm"],
        bands=[0],
    )
    layersource_dhm_hill = ows_util.WMSLayerSource(
        wms_server_url=wms_server_url_dhm,
        layernames=["DHMV_II_HILL_25cm"],
        bands=[0],
    )
    layersources = [
        layersource_dhm_skyview,
        layersource_dhm_hill,
        layersource_dhm_ortho_grey,
    ]
    image_path = ows_util.getmap_to_file(
        layersources=layersources,
        output_dir=tmp_path,
        crs=projection,
        bbox=bbox,
        size=(width_pix, height_pix),
        image_format=ows_util.FORMAT_PNG,
        # image_format_save=ows_util.FORMAT_TIFF,
        image_pixels_ignore_border=0,
        transparent=False,
        layername_in_filename=True,
    )

    assert image_path is not None
    assert image_path.exists() is True
    with rio.open(image_path) as image_file:
        # assert tuple(image_file.bounds) == bbox
        assert image_file.width == width_pix
        assert image_file.height == height_pix
